/*CAsyncSocketEx by Tim Kosse (Tim.Kosse@gmx.de)
            Version 1.2 (2003-03-28)
--------------------------------------------------------

Introduction:
-------------

CAsyncSocketEx is a replacement for the MFC class CAsyncSocket.
This class was written because CAsyncSocket is not the fastest WinSock
wrapper and it's very hard to add new functionality to CAsyncSocket
derived classes. This class offers the same functionality as CAsyncSocket.
Also, CAsyncSocketEx offers some enhancements which were not possible with
CAsyncSocket without some tricks.

How do I use it?
----------------
Basically exactly like CAsyncSocket.
To use CAsyncSocketEx, just replace all occurrences of CAsyncSocket in your
code with CAsyncSocketEx, if you did not enhance CAsyncSocket yourself in
any way, you won't have to change anything else in your code.

Why is CAsyncSocketEx faster?
-----------------------------

CAsyncSocketEx is slightly faster when dispatching notification event messages.
First have a look at the way CAsyncSocket works. For each thread that uses
CAsyncSocket, a window is created. CAsyncSocket calls WSAAsyncSelect with
the handle of that window. Until here, CAsyncSocketEx works the same way.
But CAsyncSocket uses only one window message (WM_SOCKET_NOTIFY) for all
sockets within one thread. When the window recieve WM_SOCKET_NOTIFY, wParam
contains the socket handle and the window looks up an CAsyncSocket instance
using a map. CAsyncSocketEx works differently. It's helper window uses a
wide range of different window messages (WM_USER through 0xBFFF) and passes
a different message to WSAAsyncSelect for each socket. When a message in
the specified range is received, CAsyncSocketEx looks up the pointer to a
CAsyncSocketEx instance in an Array using the index of message - WM_USER.
As you can see, CAsyncSocketEx uses the helper window in a more efficient
way, as it don't have to use the slow maps to lookup it's own instance.
Still, speed increase is not very much, but it may be noticeable when using
a lot of sockets at the same time.
Please note that the changes do not affect the raw data throughput rate,
CAsyncSocketEx only dispatches the notification messages faster.

What else does CAsyncSocketEx offer?
------------------------------------

CAsyncSocketEx offers a flexible layer system. One example is the proxy layer.
Just create an instance of the proxy layer, configure it and add it to the layer
chain of your CAsyncSocketEx instance. After that, you can connect through
proxies.
Benefit: You don't have to change much to use the layer system.
Another layer that is currently in development is the SSL layer to establish
SSL encrypted connections.

License
-------

Feel free to use this class, as long as you don't claim that you wrote it
and this copyright notice stays intact in the source files.
If you use this class in commercial applications, please send a short message
to tim.kosse@gmx.de
*/

#include "stdafx.h"
#include <string.h>
#include <stdio.h>
#if defined(WIN32)
#include <TCHAR.H>
#endif

#include "AsyncSocketEx.h"

#if defined(WIN32)
#include "AsyncSocketExLayer.h"
#endif

//#define LOG2	DLOG
#if _MSC_VER >= 1300
#define LOG2	__noop
#else
#define LOG2	0
#endif




/////////////////////////////
//Helper Window class

#define WM_SOCKETEX_NOTIFY			(WM_USER+3)
#define WM_SOCKETEX_DEAD			(WM_USER+2)
#define UM_ASYNC_GETHOSTBYNAME		(WM_USER+1)

#define IDT_CONNECTION_TIMEOUT		(WM_USER+4)

//static 
socketmap_t CSocketHelperWnd::ms_mapConnectionSocketList;
CRITICAL_SECTION CSocketHelperWnd::ms_hCriticalSection;
TCHAR CSocketHelperWnd::ms_szSocketWndClassName[256] = {0,};
BOOL CSocketHelperWnd::m_bConnectionCheckTimerActivate = FALSE;

//static 
HINSTANCE CSocketHelperWnd::ms_hInstance = NULL;

// static member
CSocketHelperWnd * CAsyncSocketEx::ms_pHelperWindow = NULL;

SOCKET_MAPPING_TABLE_ITEM * CAsyncSocketEx::ms_pSocketMappingTable = NULL;
int CAsyncSocketEx::ms_nSocketMappingTableSize = 0;
int CAsyncSocketEx::ms_nSocketMappingTable_LastIndex = -1;
int CAsyncSocketEx::ms_nSocketMappingTable_UsedItemCount = 0;
int CAsyncSocketEx::ms_nSocketMappingTable_DeadItemCount = 0;

//static
void CSocketHelperWnd::sRegisterClass(HINSTANCE hInstance)
{
	::InitializeCriticalSection(&ms_hCriticalSection);	

    WNDCLASSEX wndclass = {0};
	wndclass.cbSize=sizeof wndclass;
	
	_tcscpy(ms_szSocketWndClassName, _T("NpSessionBase:CAsyncSocketEx Helper Window"));
	while( GetClassInfoEx(hInstance, ms_szSocketWndClassName, &wndclass) )
	{
		DWORD dwError = GetLastError();
		char szMsg[512];
		_snprintf(szMsg, 512, "GetClassInfoEx return true, already registered, window class name=%s, error=%d",
			ms_szSocketWndClassName, 
			dwError );
		MessageBox(NULL, szMsg, "[CSocketHelperWnd]", MB_OK|MB_ICONERROR);
		_tcscat(ms_szSocketWndClassName, "_");
	}
	
	wndclass.style=0;
	wndclass.lpfnWndProc=WindowProc;
	wndclass.cbClsExtra=0;
	wndclass.cbWndExtra=0;
	wndclass.hInstance=hInstance;
	wndclass.hIcon=0;
	wndclass.hCursor=0;
	wndclass.hbrBackground=0;
	wndclass.lpszMenuName=0;
	wndclass.lpszClassName= ms_szSocketWndClassName;
	wndclass.hIconSm=0;
	
	if( 0 == RegisterClassEx(&wndclass) )
	{
		DWORD dwError = GetLastError();
		char szMsg[512];
		_snprintf(szMsg, 512, "RegisterClassEx fail!!, window class name=%s, error=%d",
			ms_szSocketWndClassName, 
			dwError );
		MessageBox(NULL, szMsg, "[CSocketHelperWnd]", MB_OK|MB_ICONERROR);
		return;
	}

	ms_hInstance = hInstance;
}


// static
void CSocketHelperWnd::sUnregisterClass()
{
    BOOL bRet = FALSE;
	sKillConnectionTimer();
	ms_mapConnectionSocketList.clear();

	if(NULL != ms_hInstance)
	{
		API_VERIFY(UnregisterClass( ms_szSocketWndClassName, ms_hInstance ));
		ms_hInstance = NULL;

		::DeleteCriticalSection(&ms_hCriticalSection);	
	}

	ms_szSocketWndClassName[0] = 0;
}

// static
void CSocketHelperWnd::sStartConnectionTimer()
{
	if(NULL == ms_hInstance)
		return;

    BOOL bRet = FALSE;
	if(FALSE == m_bConnectionCheckTimerActivate)
	{
		HWND hwnd = CAsyncSocketEx::ms_pHelperWindow ? 
						CAsyncSocketEx::ms_pHelperWindow->GetHwnd() : NULL;

		if(hwnd)
		{
			API_VERIFY(::KillTimer(hwnd, IDT_CONNECTION_TIMEOUT));
			API_VERIFY(0 != SetTimer(hwnd, IDT_CONNECTION_TIMEOUT, 1000, NULL));
			m_bConnectionCheckTimerActivate = TRUE;
		}
	}
}

// static
void CSocketHelperWnd::sKillConnectionTimer()
{
	if(NULL == ms_hInstance)
		return;

    BOOL bRet = FALSE;
	if(TRUE == m_bConnectionCheckTimerActivate)
	{
		HWND hwnd = CAsyncSocketEx::ms_pHelperWindow ? 
						CAsyncSocketEx::ms_pHelperWindow->GetHwnd() : NULL;

		if(hwnd)
		{
			API_VERIFY( 0 != ::KillTimer(hwnd, IDT_CONNECTION_TIMEOUT));
		}
		m_bConnectionCheckTimerActivate = FALSE;
	}
}

// static
void CSocketHelperWnd::sAddConnectionSocketToMap(CAsyncSocketEx *pSocket)
{
	if(NULL == ms_hInstance)
		return;

	if(NULL == pSocket)
		return;

	::EnterCriticalSection(&ms_hCriticalSection);

	ms_mapConnectionSocketList.insert(socketmap_t::value_type(pSocket, pSocket));

	if(ms_mapConnectionSocketList.empty() == false)
	{
		sStartConnectionTimer();
	}

	::LeaveCriticalSection(&ms_hCriticalSection);
}


// static
void CSocketHelperWnd::sRemoveConnectionSocketFromMap(CAsyncSocketEx *pSocket)
{
	if(NULL == ms_hInstance)
		return;

	if(NULL == pSocket)
		return;

	if(ms_mapConnectionSocketList.empty())
		return;

	::EnterCriticalSection(&ms_hCriticalSection);
		
	socketmap_t::iterator it = ms_mapConnectionSocketList.find( pSocket );

	if( it != ms_mapConnectionSocketList.end() )
	{
		ms_mapConnectionSocketList.erase(it);

		if(ms_mapConnectionSocketList.empty())
		{
			sKillConnectionTimer();
		}
	}

	::LeaveCriticalSection(&ms_hCriticalSection);
}

CSocketHelperWnd::CSocketHelperWnd()
{
#if defined(WIN32)
	//Create window

	m_hWnd=CreateWindow( ms_szSocketWndClassName, _T("CAsyncSocketEx Helper Window"), 
		WS_POPUP, 0, 0, 0, 0,
		0, 
		(HMENU)0, 
		ms_hInstance, 
		0);
	ASSERT(m_hWnd);
	SetWindowLong(m_hWnd, GWL_USERDATA, (LONG)this);
#endif

};

CSocketHelperWnd::~CSocketHelperWnd()
{
#if defined(WIN32)
    BOOL bRet = FALSE;
	//Destroy window
	if (m_hWnd)
	{
		API_VERIFY(DestroyWindow(m_hWnd));
		m_hWnd=0;
	}
#endif	
	m_mapAsyncGetHostByName.clear();
}


// static
//Processes event notifications sent by the sockets or the layers

LRESULT CALLBACK CSocketHelperWnd::WindowProc( HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam )
{
	LRESULT ret = DefWindowProc(hWnd, message, wParam, lParam);
    BOOL bRet = FALSE;

	if( message == WM_TIMER )
	{
		CSocketHelperWnd *pWnd= (CSocketHelperWnd *)GetWindowLong(hWnd, GWL_USERDATA);
        ATLASSERT(pWnd != NULL); //can assert this?
		if(NULL != pWnd)
		{
			::EnterCriticalSection(&ms_hCriticalSection);
			
			socketmap_t::iterator it1 = ms_mapConnectionSocketList.begin();
			for (; it1 != ms_mapConnectionSocketList.end(); ++it1)
			{
				if(--it1->second->m_nConnectionTimeout <= 0)
				{
					// KEH_CONNECTION_TIMEOUT
					// TIMEOUT Ǿ!!
					it1->second->OnConnect(WSAETIMEDOUT);

					if(INVALID_SOCKET != it1->second->m_hSocket)
					{
						if(NULL != pWnd->GetHwnd())
							it1->second->TerminateAsyncSelect();
						it1->second->Close();
					}

					// ʿ ٷ Ѵ.
					ms_mapConnectionSocketList.erase(it1++);
				}
			}
			
			//  for ȿ erase ֱ 
			// sRemoveConnectionSocketFromMap Լ   .
			//  ʿ ϴ ڵ尡 Ƿ
			//  ִ ⼭ üũ.
			if(ms_mapConnectionSocketList.empty())
			{
				sKillConnectionTimer();
			}

			::LeaveCriticalSection(&ms_hCriticalSection);
		}
	}
#if defined(WIN32)
	else if( message == WM_USER ) //Notification event sent by a layer
	{
		//Verify parameters, lookup socket and notification message
		if (!wParam)
			return 0;
		CAsyncSocketEx *pSocket=(CAsyncSocketEx *)wParam;
		CAsyncSocketExLayer::t_LayerNotifyMsg *pMsg=(CAsyncSocketExLayer::t_LayerNotifyMsg *)lParam;
		if (pSocket->GetSocketHandle() == INVALID_SOCKET)
		{
			delete pMsg;
			return 0;
		}
		int nEvent=pMsg->lEvent&0xFFFF;
		int nErrorCode=pMsg->lEvent>>16;
		
		//Dispatch to layer
		if (pMsg->pLayer)
			pMsg->pLayer->CallEvent(nEvent, nErrorCode);
		else
		{
			//Dispatch to socket class
			switch (nEvent)
			{
			case FD_READ:
				if (pSocket->m_lEvent & FD_READ)
					pSocket->OnReceive(nErrorCode, -1);
				break;
			case FD_FORCEREAD:
				if (pSocket->m_lEvent & FD_FORCEREAD)
					pSocket->OnReceive(nErrorCode, -1);
				break;
			case FD_WRITE:
				if (pSocket->m_lEvent & FD_WRITE)
					pSocket->OnSend(nErrorCode);
				break;
			case FD_CONNECT:
				if (pSocket->m_lEvent & FD_CONNECT)
				{
					CSocketHelperWnd::sRemoveConnectionSocketFromMap(pSocket);
					pSocket->OnConnect(nErrorCode);
				}
				break;
			case FD_ACCEPT:
				if (pSocket->m_lEvent & FD_ACCEPT)
					pSocket->OnAccept(nErrorCode);
				break;
			case FD_CLOSE:
				if (pSocket->m_lEvent & FD_CLOSE)
					pSocket->OnClose(nErrorCode);
				break;
			}
		}
		delete pMsg;
		return 0;
	}
#endif
	else if( message == UM_ASYNC_GETHOSTBYNAME )
	{
		//WSAAsyncGetHostByName reply		

		//Verify parameters
		ASSERT(hWnd);
		CSocketHelperWnd *pWnd=(CSocketHelperWnd *)GetWindowLong(hWnd, GWL_USERDATA);
		ASSERT(pWnd);
		
		HANDLE hTask = (HANDLE)wParam;
//#define WSAGETASYNCERROR(lParam)            HIWORD(lParam)
//#define WSAGETASYNCBUFLEN(lParam)           LOWORD(lParam)

		map<HANDLE, CAsyncSocketEx *>::iterator it = pWnd->m_mapAsyncGetHostByName.find( hTask );
		//ASSERT( it != pWnd->m_mapAsyncGetHostByName.end() );	//  ޽ dispatch Ǳ  ̹  ų  ۾ ҵǾٸ hTask  شϴ map  . [taekwonv]
		if( it == pWnd->m_mapAsyncGetHostByName.end() )
			return 0;

		CAsyncSocketEx * pSocket = it->second;		
		pWnd->m_mapAsyncGetHostByName.erase( it );
		pSocket->m_hAsyncGetHostByNameHandle=0;

//		int nErrorCode = lParam >> 16;
//		int len = lParam % 0xFFFF;
		int nErrorCode = WSAGETASYNCERROR(lParam);
		if( nErrorCode == WSAENOBUFS )
		{
			int nRequireBuffLen = WSAGETASYNCBUFLEN(lParam);
			
			// buffer realloc
			delete [] pSocket->m_pAsyncGetHostByNameBuffer;
			pSocket->m_pAsyncGetHostByNameBuffer= new char[nRequireBuffLen];
			
			pSocket->m_hAsyncGetHostByNameHandle = WSAAsyncGetHostByName(pWnd->GetHwnd(), UM_ASYNC_GETHOSTBYNAME, pSocket->m_pszAsyncGetHostByNameDomain, pSocket->m_pAsyncGetHostByNameBuffer, nRequireBuffLen);
			if (!pSocket->m_hAsyncGetHostByNameHandle)
			{
				int nErrorCode = WSAGetLastError();
				CSocketHelperWnd::sRemoveConnectionSocketFromMap(pSocket);
                pSocket->OnConnect(nErrorCode);
				return 0;
			}

			pWnd->m_mapAsyncGetHostByName.insert( 
				map<HANDLE, CAsyncSocketEx *>::value_type(pSocket->m_hAsyncGetHostByNameHandle, pSocket) );

			WSASetLastError(WSAEWOULDBLOCK);
			return 0;
		}

		if( nErrorCode )
		{
			if( pSocket->m_pAsyncGetHostByNameBuffer )
			{
				delete [] pSocket->m_pAsyncGetHostByNameBuffer;
				pSocket->m_pAsyncGetHostByNameBuffer=NULL;
			}

			if( pSocket->m_pszAsyncGetHostByNameDomain )
			{
				free( pSocket->m_pszAsyncGetHostByNameDomain );
				pSocket->m_pszAsyncGetHostByNameDomain = NULL;
				pSocket->m_nAsyncGetHostByNamePort = 0;
			}

			// pSocket ڵ鷯 ȣ  pSocket  ȿ   ִ.  (:CASyncSocketEx ڵ鷯 delete this  ϴ )
			//  pSocket ڵ鷯 ȣ Ŀ pSocket  ݵ ؼ ȵ. [taekwonv]
			CSocketHelperWnd::sRemoveConnectionSocketFromMap(pSocket);
			pSocket->OnConnect(nErrorCode);			

			return 0;
		}
		
		SOCKADDR_IN sockAddr;
		memset(&sockAddr,0,sizeof(sockAddr));
		sockAddr.sin_family=AF_INET;
		sockAddr.sin_addr.s_addr = ((LPIN_ADDR)((LPHOSTENT)pSocket->m_pAsyncGetHostByNameBuffer)->h_addr)->s_addr;
		
		sockAddr.sin_port = htons(pSocket->m_nAsyncGetHostByNamePort);
		
		API_VERIFY(pSocket->Connect((SOCKADDR*)&sockAddr, sizeof(sockAddr)));

		if( pSocket->m_pAsyncGetHostByNameBuffer )
		{
			delete [] pSocket->m_pAsyncGetHostByNameBuffer;
			pSocket->m_pAsyncGetHostByNameBuffer=NULL;
		}
		if( pSocket->m_pszAsyncGetHostByNameDomain )
		{
			free( pSocket->m_pszAsyncGetHostByNameDomain );
			pSocket->m_pszAsyncGetHostByNameDomain = NULL;
			pSocket->m_nAsyncGetHostByNamePort = 0;
		}

		if( !bRet)
		{
			DWORD dwError = GetLastError();
			if( dwError != WSAEWOULDBLOCK )
			{
				CSocketHelperWnd::sRemoveConnectionSocketFromMap(pSocket);
				pSocket->OnConnect( dwError );
			}
		}	

		return 0;
	}
	else if( message == WM_SOCKETEX_DEAD )
	{
		CAsyncSocketEx::sRemoveSocketDeadMark(wParam);
	}
	else if( message >= WM_SOCKETEX_NOTIFY && message < WM_SOCKETEX_NOTIFY + CAsyncSocketEx::ms_nSocketMappingTableSize )
	{
		CAsyncSocketEx::sDoCallBack(message - WM_SOCKETEX_NOTIFY,  wParam, lParam );
	}
	
	return ret ;
}


// static
int PASCAL CAsyncSocketEx::sFind_SMT_NextAvailableIndex()
{
	int nNewIndex = ms_nSocketMappingTable_LastIndex;

	while( 1 )
	{
		nNewIndex++;
		
		if( nNewIndex >= ms_nSocketMappingTableSize )
			nNewIndex = 0;

		if( nNewIndex == ms_nSocketMappingTable_LastIndex )
		{
			ASSERT( false );
			return -1;
		}

		ASSERT( nNewIndex >= 0 && nNewIndex < ms_nSocketMappingTableSize);
		if( ms_pSocketMappingTable[nNewIndex].eState == eSS_NOT_USED )
		{
			return nNewIndex;
		}
	}
}

//static 
void CAsyncSocketEx::sDumpMappingTable()
{
	/*
	int nTotalLen = 120 + ms_nSocketMappingTableSize * (3 + 1 + 1 + 1 + 1+2+ 1 +2+1+2+1+2+1+3+1+ 1 + 8 + 5);
	char * pDumpLog = (char*)malloc( nTotalLen );
	int nWritten = 0;

	SYSTEMTIME st;
	GetLocalTime( &st );
	nWritten += sprintf( pDumpLog, "[%02d %02d:%02d:%02d.%03d] CAsyncSocketEx handle mapping table dump\r\n"
		"\t dead item count=%d, last index=%d, used item count=%d\r\n", 
		st.wDay, 
		st.wHour,
		st.wMinute,
		st.wSecond,
		st.wMilliseconds,

		ms_nSocketMappingTable_DeadItemCount,
		ms_nSocketMappingTable_LastIndex,
		ms_nSocketMappingTable_UsedItemCount );

	for( int i=0 ; i < ms_nSocketMappingTableSize ; i++ )
	{
		nWritten += _snprintf(pDumpLog+nWritten, nTotalLen-nWritten, "%03d %1d [%02d %02d:%02d:%02d.%03d] %d\r\n",
			i,
			ms_pSocketMappingTable[i].eState,
			ms_pSocketMappingTable[i].last_state_changed_st.wDay, 
			ms_pSocketMappingTable[i].last_state_changed_st.wHour,
			ms_pSocketMappingTable[i].last_state_changed_st.wMinute,
			ms_pSocketMappingTable[i].last_state_changed_st.wSecond,
			ms_pSocketMappingTable[i].last_state_changed_st.wMilliseconds,
			ms_pSocketMappingTable[i].nDeadCount );
	}

	g_log.raw_write( pDumpLog, nWritten );

	free( pDumpLog );
*/
}


// static
bool PASCAL CAsyncSocketEx::sAttachHandle(CAsyncSocketEx* pSocket)
{
//	IK_TRACE("[ENTER] CAsyncSocketEx::AttachHandle(pSocket=%p)\n", pSocket);
	ASSERT( pSocket->m_nSMT_Index == -1 );
	
	if( ms_nSocketMappingTable_UsedItemCount == 0 && ms_nSocketMappingTable_DeadItemCount == 0 )
	{
		ASSERT( ms_nSocketMappingTableSize == 0 );
		ASSERT( ms_pSocketMappingTable == NULL );

		ms_nSocketMappingTableSize = DEFAULT_SMT_ITEM_COUNT;
		ms_pSocketMappingTable = new SOCKET_MAPPING_TABLE_ITEM[ms_nSocketMappingTableSize];
		ZeroMemory( ms_pSocketMappingTable, sizeof(SOCKET_MAPPING_TABLE_ITEM) * ms_nSocketMappingTableSize);
		
		ASSERT( ms_nSocketMappingTable_LastIndex == -1 );	
		ms_nSocketMappingTable_LastIndex = -1;
		
		ASSERT(ms_pHelperWindow == NULL);	
		if( ms_pHelperWindow == NULL )
		{
			ms_pHelperWindow = new CSocketHelperWnd();
			ASSERT(ms_pHelperWindow->GetHwnd() != NULL);
		}
	}
	
	if( ms_nSocketMappingTable_UsedItemCount + ms_nSocketMappingTable_DeadItemCount >= ms_nSocketMappingTableSize )
	{
		char szMsg[1024];
		_snprintf(szMsg, 1024, "ִ 밡   ʰ Ͽϴ. (1) UsedItemCount=%d, DeadItemCount=%d", 
			ms_nSocketMappingTable_UsedItemCount,
			ms_nSocketMappingTable_DeadItemCount);

		sDumpMappingTable();

		MessageBox(NULL, szMsg, "[CAsyncSocketEx]", MB_OK | MB_ICONERROR );
		
		sDumpMappingTable();

		return false;
	}
	
	pSocket->m_nSMT_Index = sFind_SMT_NextAvailableIndex();
	if( pSocket->m_nSMT_Index == -1 )
	{
		char szMsg[1024];
		_snprintf(szMsg, 1024, "ִ 밡   ʰ Ͽϴ. (2) UsedItemCount=%d, DeadItemCount=%d", 
			ms_nSocketMappingTable_UsedItemCount,
			ms_nSocketMappingTable_DeadItemCount);

		sDumpMappingTable();

		MessageBox(NULL, szMsg, "[CAsyncSocketEx]", MB_OK | MB_ICONERROR );
		
		sDumpMappingTable();

		return false;
	}

	if( pSocket->m_nSMT_Index < 0 || pSocket->m_nSMT_Index >= ms_nSocketMappingTableSize )
	{
		char szMsg[1024];
		_snprintf(szMsg, 1024, "FATAL ERROR!!\r\n pSocket->m_nSMT_Index=%d, UsedItemCount=%d, DeadItemCount=%d", 
			pSocket->m_nSMT_Index,
			ms_nSocketMappingTable_UsedItemCount,
			ms_nSocketMappingTable_DeadItemCount);

		sDumpMappingTable();

		MessageBox(NULL, szMsg, "[CAsyncSocketEx]", MB_OK | MB_ICONERROR );
		
		sDumpMappingTable();
	}
	
	ms_nSocketMappingTable_LastIndex = pSocket->m_nSMT_Index;
	ms_nSocketMappingTable_UsedItemCount++;
	
	ASSERT( ms_pSocketMappingTable[pSocket->m_nSMT_Index].eState == eSS_NOT_USED );
	ms_pSocketMappingTable[pSocket->m_nSMT_Index].eState = eSS_USED;
//	GetLocalTime( &ms_pSocketMappingTable[pSocket->m_nSMT_Index].last_state_changed_st );
	ms_pSocketMappingTable[pSocket->m_nSMT_Index].pSocket = pSocket;
//	IK_TRACE("[LEAVE]CAsyncSocketEx::AttachHandle(pSocket=%p), m_nSMT_Index=%d\n", pSocket, pSocket->m_nSMT_Index);

	return true;
}

//static 
void PASCAL CAsyncSocketEx::sAddSocketDeadMark(int nSMT_Index)
{
	if( nSMT_Index < 0 || nSMT_Index >= ms_nSocketMappingTableSize )
	{
		ASSERT( false );
		return;
	}

	if( ms_pHelperWindow == NULL )
		return;
	
	if( ms_pSocketMappingTable[nSMT_Index].eState == eSS_NOT_USED )
	{
		ms_nSocketMappingTable_DeadItemCount++;

		ms_pSocketMappingTable[nSMT_Index].eState = eSS_DEAD;
//		GetLocalTime( &ms_pSocketMappingTable[nSMT_Index].last_state_changed_st );
		ms_pSocketMappingTable[nSMT_Index].nDeadCount = 1;
	}
	else if( ms_pSocketMappingTable[nSMT_Index].eState == eSS_DEAD )
	{
		ms_pSocketMappingTable[nSMT_Index].nDeadCount++;
	}
	else
	{
		ASSERT( false );
	}
}

// static
void PASCAL CAsyncSocketEx::sDetachHandle(int nSMT_Index)
{
	if( nSMT_Index < 0 || nSMT_Index >= ms_nSocketMappingTableSize )
	{
		ASSERT( false );
		return;
	}
	
	ASSERT( ms_nSocketMappingTable_UsedItemCount > 0 );
	ASSERT( ms_pSocketMappingTable[nSMT_Index].eState == eSS_USED );
	if( ms_pSocketMappingTable[nSMT_Index].eState != eSS_USED )
		return;

	ms_pSocketMappingTable[nSMT_Index].eState = eSS_NOT_USED;
	ms_pSocketMappingTable[nSMT_Index].pSocket = NULL;

	ms_nSocketMappingTable_UsedItemCount--;
	if( ms_nSocketMappingTable_LastIndex == nSMT_Index )
		ms_nSocketMappingTable_LastIndex--;

	if( ms_nSocketMappingTable_UsedItemCount == 0 )
	{
		ASSERT( ms_pHelperWindow != NULL);
		ASSERT( ms_pHelperWindow->GetHwnd() != NULL);

		delete ms_pHelperWindow;
		ms_pHelperWindow = NULL;

		ms_nSocketMappingTable_LastIndex = -1;
		ms_nSocketMappingTable_DeadItemCount = 0;

		delete [] ms_pSocketMappingTable;
		ms_pSocketMappingTable = NULL;
		ms_nSocketMappingTableSize = 0;
	}
}

// static
void PASCAL CAsyncSocketEx::sRemoveSocketDeadMark(int nSMT_Index)
{
    if( nSMT_Index < 0 || nSMT_Index >= ms_nSocketMappingTableSize )
	{
		ASSERT( false );
		return;
	}

	ASSERT(ms_pSocketMappingTable[nSMT_Index].eState == eSS_DEAD);

	if( ms_pSocketMappingTable[nSMT_Index].eState != eSS_DEAD )
		return;

	ms_pSocketMappingTable[nSMT_Index].nDeadCount--;
	if( ms_pSocketMappingTable[nSMT_Index].nDeadCount <= 0 )
	{
		ASSERT( ms_pSocketMappingTable[nSMT_Index].nDeadCount == 0 );

		ms_pSocketMappingTable[nSMT_Index].eState = eSS_NOT_USED;
//		GetLocalTime( &ms_pSocketMappingTable[nSMT_Index].last_state_changed_st );
		ms_pSocketMappingTable[nSMT_Index].nDeadCount = 0;

		ms_nSocketMappingTable_DeadItemCount--;
		ASSERT( ms_nSocketMappingTable_DeadItemCount >= 0 );
	}
}

// static
void PASCAL CAsyncSocketEx::sKillSocket(int nSMT_Index, CAsyncSocketEx* pSocket)
{
	if( nSMT_Index < 0 || nSMT_Index >= ms_nSocketMappingTableSize )
	{
		ASSERT( false );
	}

	ASSERT(ms_pSocketMappingTable[nSMT_Index].eState == eSS_USED);

	sDetachHandle(nSMT_Index);

	if( ms_pHelperWindow != NULL )
	{
		ASSERT(ms_pSocketMappingTable[nSMT_Index].eState == eSS_NOT_USED);

		::PostMessage(ms_pHelperWindow->GetHwnd(), WM_SOCKETEX_DEAD,
			(WPARAM)nSMT_Index, 0L);
		sAddSocketDeadMark(nSMT_Index);
	}
}

// static
void PASCAL CAsyncSocketEx::sDoCallBack(int nSMT_Index, WPARAM wParam, LPARAM lParam)
{
	if (wParam == 0 && lParam == 0)
		return;

	int nErrorCode	= WSAGETSELECTERROR(lParam);
	int nEvent		= WSAGETSELECTEVENT(lParam);

	if( nSMT_Index < 0 || nSMT_Index >= ms_nSocketMappingTableSize )
	{
		char szMsg[1024];
		_snprintf(szMsg, 1024, "CAsyncSocketEx::sDoCallBack(), nSMT_Index  . nSMT_Index=%d\r\n"
			"nEvent=%d, nErrorCode=%d, wParam=%d",
			nSMT_Index,
			nEvent,
			nErrorCode,
			wParam );

		DLOG("%s", szMsg);

		sDumpMappingTable();

		//MessageBox(NULL, szMsg, "[CAsyncSocketEx] ERROR", MB_OK | MB_ICONERROR );

		return;
	}

	CAsyncSocketEx * pSocket = NULL;
	if( ms_pSocketMappingTable[nSMT_Index].eState == eSS_USED )
		pSocket = ms_pSocketMappingTable[nSMT_Index].pSocket;
	else if( ms_pSocketMappingTable[nSMT_Index].eState == eSS_DEAD )
	{
		// socket close  -- 
		return;
	}
	else
	{
#ifdef _DEBUG
		char szMsg[1024];
		_snprintf(szMsg, 1024, "CAsyncSocketEx::sDoCallBack(), CAsyncSocketEx::LookupHandle() FAIL!!\r\n"
			"nEvent=%d, nErrorCode=%d, socket_handle(wParam)=%d",
			nEvent,
			nErrorCode,
			wParam );

		DLOG("%s", szMsg);
			
		sDumpMappingTable();

		MessageBox(NULL, szMsg, "[CAsyncSocketEx] ERROR", MB_OK | MB_ICONERROR);
#endif
		return;
	}


	if( pSocket->m_bIsTcpListenSocket )
	{
		if( pSocket->m_hSocket != (SOCKET)wParam )
		{
			// ̰ װ ƴϴ.
			//
			// accept() Լ Ҹ , accept()    Ͽ ؼ 
			// notify message  ޽ ť . --> ̶, notify message  accept Ҷ,
			// listen  ϰ ִ Ͽ ؼ  message ID  PostMessage  Ǳ  ̷
			//  ߻ϴ ̴. , ׳ ϸ, accept() Ŀ Ҹ WSAAsyncSelect()  ؼ 
			// (츮 ϴ) socket notify message   ,, ⼭ ׳ Ѵ.
			
			return;
		}
	}
	else
	{
		if( pSocket->m_hSocket != (SOCKET)wParam )
		{
			//  ,,
			// windows 98  ߻ϴ° δ. -->  process 踦 Ѿ ޵Ǿ 쿡 ߻
			// 
			DLOG("pSocket->m_bIsTcpListenSocket == false, pSocket->m_hSocket[%d] != (SOCKET)wParam[%d]\n", 
				pSocket->m_hSocket, wParam );
		}
	}
#if defined(WIN32)
	if (pSocket->m_pFirstLayer)
	{
		if (nEvent==FD_READ)
		{
			DWORD nBytes;
			if (!pSocket->IOCtl(FIONREAD, &nBytes))
				nErrorCode = WSAGetLastError();
			if (nBytes != 0 || nErrorCode != 0)
				pSocket->m_pLastLayer->CallEvent(nEvent, nErrorCode);
		}
		else
			pSocket->m_pLastLayer->CallEvent(nEvent, nErrorCode);
		return;
	}
#endif
	switch( nEvent )
	{
	case FD_READ:
		{
			DWORD nBytes;
			if (!pSocket->IOCtl(FIONREAD, &nBytes))
				nErrorCode = WSAGetLastError();
			if (nBytes != 0 || nErrorCode != 0)
				pSocket->OnReceive(nErrorCode, nBytes);
		}
		break;
	case FD_WRITE:
		pSocket->OnSend(nErrorCode);
		break;
	case FD_OOB:
		pSocket->OnOutOfBandData(nErrorCode);
		break;
	case FD_ACCEPT:
		pSocket->OnAccept(nErrorCode);
		break;
	case FD_CONNECT:
		CSocketHelperWnd::sRemoveConnectionSocketFromMap(pSocket);
		pSocket->OnConnect(nErrorCode);
		break;
	case FD_CLOSE:
		pSocket->OnClose(nErrorCode);
		break;
	}
}

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CAsyncSocketEx::CAsyncSocketEx()
{
	m_bRequestConnect = false;
	m_bIsTcpListenSocket = false;
	m_nSMT_Index = -1;
	m_hSocket = INVALID_SOCKET;
	m_bNotifyClosed = false;
	
	m_pFirstLayer = 0;
	m_pLastLayer = 0;

	m_nConnectionTimeout = -1;

	m_pszAsyncGetHostByNameDomain = NULL;
	m_nAsyncGetHostByNamePort = 0;
	m_pAsyncGetHostByNameBuffer = NULL;
	m_hAsyncGetHostByNameHandle = NULL;
}

CAsyncSocketEx::~CAsyncSocketEx()
{
	DLOG("CAsyncSocketEx::~CAsyncSocketEx(), this=%p, m_hSocket=%d\n"
		"ms_nSocketMappingTable_LastIndex=%d\n"
		"ms_nSocketMappingTable_UsedItemCount=%d\n" 
		"ms_nSocketMappingTable_DeadItemCount=%d\n", 
		this, m_hSocket, 
		ms_nSocketMappingTable_LastIndex,
		ms_nSocketMappingTable_UsedItemCount,
		ms_nSocketMappingTable_DeadItemCount );

	Close();

	CSocketHelperWnd::sRemoveConnectionSocketFromMap(this);
}


BOOL CAsyncSocketEx::Socket(int nSocketType, long lEvent,
	int nProtocolType, int nAddressFormat)
{
    int rc = SOCKET_ERROR;
	ASSERT(m_hSocket == INVALID_SOCKET);

	m_hSocket = socket(nAddressFormat, nSocketType, nProtocolType);
	m_bNotifyClosed = false;
	if (m_hSocket != INVALID_SOCKET)
	{
		if( !sAttachHandle( this ) )
		{
			NET_VERIFY(closesocket( m_hSocket ));
			m_hSocket = INVALID_SOCKET;
			return FALSE;
		}

		return AsyncSelect(lEvent);
	}
    else
    {
        NET_VERIFY(SOCKET_ERROR); //make sure ASSERT and display error in debug version
    }
	return FALSE;	
}

BOOL CAsyncSocketEx::Create( UINT nSocketPort /*=0*/, int nSocketType /*=SOCK_STREAM*/, long lEvent /*=FD_READ | FD_WRITE | FD_OOB | FD_ACCEPT | FD_CONNECT | FD_CLOSE*/, LPCTSTR lpszSocketAddress /*=NULL*/ )
{
	ASSERT(GetSocketHandle()==INVALID_SOCKET);

	//Close the socket, although this should not happen
	if (GetSocketHandle() != INVALID_SOCKET)
	{
		WSASetLastError(WSAEALREADY);
		return FALSE;
	}
#if defined(WIN32)
	if (m_pFirstLayer)
		return m_pFirstLayer->Create(nSocketPort, nSocketType, lEvent, lpszSocketAddress);
	else
	{
#endif
		if( Socket( nSocketType, lEvent ) )
		{
			if (Bind(nSocketPort, lpszSocketAddress))
				return TRUE;
			
			int nResult = GetLastError();
			Close();
			WSASetLastError(nResult);
		}
#if defined(WIN32)
	}
#endif
	
	return FALSE;
}



void CAsyncSocketEx::OnReceive(int nErrorCode, int nBytes)
{
}

void CAsyncSocketEx::OnSend(int nErrorCode)
{
}

void CAsyncSocketEx::OnConnect(int nErrorCode)
{
}

void CAsyncSocketEx::OnAccept(int nErrorCode)
{
}

void CAsyncSocketEx::OnOutOfBandData(int nErrorCode)
{
}

void CAsyncSocketEx::OnClose(int nErrorCode)
{
}

BOOL CAsyncSocketEx::Bind(UINT nSocketPort, LPCTSTR lpszSocketAddress)
{
    int rc = SOCKET_ERROR;
//	USES_CONVERSION;

	SOCKADDR_IN sockAddr;
	memset(&sockAddr,0,sizeof(sockAddr));

//	LPSTR lpszAscii = T2A((LPTSTR)lpszSocketAddress);
	LPSTR lpszAscii = (LPTSTR)lpszSocketAddress;

	sockAddr.sin_family = AF_INET;

	if (lpszAscii == NULL)
		sockAddr.sin_addr.s_addr = htonl(INADDR_ANY);
	else
	{
		DWORD lResult = inet_addr(lpszAscii);
		if (lResult == INADDR_NONE)
		{
			WSASetLastError(WSAEINVAL);
            NET_VERIFY(SOCKET_ERROR); //make sure ASSERT here
			return FALSE;
		}
		sockAddr.sin_addr.s_addr = lResult;
	}

	sockAddr.sin_port = htons((u_short)nSocketPort);

	return Bind((SOCKADDR*)&sockAddr, sizeof(sockAddr));
}

BOOL CAsyncSocketEx::Bind(const SOCKADDR* lpSockAddr, int nSockAddrLen)
{
    int rc = SOCKET_ERROR;

	NET_VERIFY(bind(m_hSocket, lpSockAddr, nSockAddrLen));
    return (rc != SOCKET_ERROR);
}

void CAsyncSocketEx::Close()
{
    int rc = 0;
#if defined(WIN32)
	if( m_pFirstLayer )
		m_pFirstLayer->Close();
#endif
	if( m_hAsyncGetHostByNameHandle )
	{
		ms_pHelperWindow->m_mapAsyncGetHostByName.erase( m_hAsyncGetHostByNameHandle );
		NET_VERIFY(WSACancelAsyncRequest(m_hAsyncGetHostByNameHandle));
		m_hAsyncGetHostByNameHandle = NULL;
	}

	if( m_pszAsyncGetHostByNameDomain )
	{
		free( m_pszAsyncGetHostByNameDomain );
		m_pszAsyncGetHostByNameDomain = NULL;
	}
	m_nAsyncGetHostByNamePort = 0;

	if( m_pAsyncGetHostByNameBuffer )
	{
		delete [] m_pAsyncGetHostByNameBuffer;
		m_pAsyncGetHostByNameBuffer = NULL;
	}

	if( m_hSocket != INVALID_SOCKET )
	{
		NET_VERIFY(closesocket(m_hSocket));
		sKillSocket(m_nSMT_Index, this);
		m_nSMT_Index = -1;
		m_hSocket = INVALID_SOCKET;
	
		CSocketHelperWnd::sRemoveConnectionSocketFromMap(this);

		m_bIsTcpListenSocket = false;
		m_bNotifyClosed = false;
	}

	RemoveAllLayers();
}

int CAsyncSocketEx::Receive(void* lpBuf, int nBufLen, int nFlags /*=0*/)
{
    int rc = SOCKET_ERROR;
#if defined(WIN32)
	if (m_pFirstLayer)
    {
		NET_VERIFY(m_pFirstLayer->Receive(lpBuf, nBufLen, nFlags));
    }
	else
#endif
    {
		NET_VERIFY(recv(m_hSocket, (LPSTR)lpBuf, nBufLen, nFlags));
    }
    return rc;
}

//	int ReceiveFrom(void* lpBuf, int nBufLen,
//		CString& rSocketAddress, UINT& rSocketPort, int nFlags = 0);

int CAsyncSocketEx::ReceiveFrom(void* lpBuf, int nBufLen,
		SOCKADDR* lpSockAddr, int* lpSockAddrLen, int nFlags)
{
    int rc = SOCKET_ERROR;
	NET_VERIFY(recvfrom(m_hSocket, (LPSTR)lpBuf, nBufLen, nFlags, lpSockAddr, lpSockAddrLen));
    return rc;
}

int CAsyncSocketEx::Send(const void* lpBuf, int nBufLen, int nFlags /*=0*/)
{
    int rc = 0;
#if defined(WIN32)
	if (m_pFirstLayer)
    {
		NET_VERIFY(m_pFirstLayer->Send(lpBuf, nBufLen, nFlags));
    }
	else
#endif
    {
		NET_VERIFY(send(m_hSocket, (LPSTR)lpBuf, nBufLen, nFlags));
    }
    return rc;
}

//	int SendTo(const void* lpBuf, int nBufLen,
//		UINT nHostPort, LPCTSTR lpszHostAddress = NULL, int nFlags = 0);

int CAsyncSocketEx::SendTo(const void * lpBuf, int nBufLen,
		const SOCKADDR* lpSockAddr, int nSockAddrLen, int nFlags)
{
    int rc = SOCKET_ERROR;
	NET_VERIFY(sendto(m_hSocket, (LPSTR)lpBuf, nBufLen, nFlags, lpSockAddr, nSockAddrLen));
    return rc;
}

BOOL CAsyncSocketEx::InitSocket()
{
	WSADATA wsaData;
	WORD wVersionRequested = MAKEWORD(2, 2);
	
	int nResult = WSAStartup(wVersionRequested, &wsaData);
	if (nResult != 0)
    {
        REPORT_ERROR_INFO(FTL::CFNetErrorInfo, nResult, WSAStartup);
		return FALSE;
    }
	if (LOBYTE(wsaData.wVersion) != 2 || HIBYTE(wsaData.wVersion) != 2)
	{
        ASSERT(FALSE);
		WSACleanup();
		return FALSE;
	}
	
	return TRUE;
}

BOOL CAsyncSocketEx::Connect(LPCTSTR lpszHostAddress, UINT nHostPort, int nTimeOut /*= -1*/ )
{
	m_bRequestConnect = true;

#if defined(WIN32)
	if (m_pFirstLayer)
    {
		return m_pFirstLayer->Connect(lpszHostAddress, nHostPort, nTimeOut);
    }
	else
	{
#endif
		ASSERT(lpszHostAddress != NULL);

		SOCKADDR_IN sockAddr;
		memset(&sockAddr,0,sizeof(sockAddr));

		sockAddr.sin_family = AF_INET;
		sockAddr.sin_addr.s_addr = inet_addr(lpszHostAddress);

		if (sockAddr.sin_addr.s_addr == INADDR_NONE)
		{
			ASSERT( m_pAsyncGetHostByNameBuffer == NULL );
			ASSERT( m_hAsyncGetHostByNameHandle == NULL );
			if( m_pAsyncGetHostByNameBuffer )
				delete [] m_pAsyncGetHostByNameBuffer;
			m_pAsyncGetHostByNameBuffer = new char[MAXGETHOSTSTRUCT];

			if( m_pszAsyncGetHostByNameDomain )
				free( m_pszAsyncGetHostByNameDomain );
			m_pszAsyncGetHostByNameDomain = strdup( lpszHostAddress );
			m_nAsyncGetHostByNamePort=nHostPort;

			m_hAsyncGetHostByNameHandle=WSAAsyncGetHostByName(GetHelperWindowHandle(), UM_ASYNC_GETHOSTBYNAME, m_pszAsyncGetHostByNameDomain, m_pAsyncGetHostByNameBuffer, MAXGETHOSTSTRUCT);
			if (!m_hAsyncGetHostByNameHandle)
				return FALSE;

			ms_pHelperWindow->m_mapAsyncGetHostByName.insert( 
				map<HANDLE, CAsyncSocketEx *>::value_type(m_hAsyncGetHostByNameHandle, this) );

			WSASetLastError(WSAEWOULDBLOCK);
			return TRUE;
		}


		sockAddr.sin_port = htons((u_short)nHostPort);
		if( sockAddr.sin_port == 0 )
		{
			WSACleanup();
			CAsyncSocketEx::InitSocket();
		}

		return Connect((SOCKADDR*)&sockAddr, sizeof(sockAddr), nTimeOut);
#if defined(WIN32)
	}
#endif
}

BOOL CAsyncSocketEx::Connect( const SOCKADDR* lpSockAddr, int nSockAddrLen, int nTimeOut /*= -1*/ )
{
    int rc = SOCKET_ERROR;
#if defined(WIN32)
	if (m_pFirstLayer)
    {
        return m_pFirstLayer->Connect(lpSockAddr, nSockAddrLen, nTimeOut);
    }
	else
	{
#endif
		// KEH_CONNECTION_TIMEOUT
		// TODO
		m_nConnectionTimeout = nTimeOut;
		if(m_nConnectionTimeout > 0)
		{
			CSocketHelperWnd::sAddConnectionSocketToMap(this);
		}
		NET_VERIFY_EXCEPT1(connect(m_hSocket, lpSockAddr, nSockAddrLen), WSAEWOULDBLOCK);
        return (SOCKET_ERROR != rc);
#if defined(WIN32)
	}
#endif
}


#ifdef _AFX
BOOL CAsyncSocketEx::GetPeerName( CString& rPeerAddress, UINT& rPeerPort )
{
	if (m_pFirstLayer)
		return m_pFirstLayer->GetPeerName(rPeerAddress, rPeerPort);

	SOCKADDR_IN sockAddr;
	memset(&sockAddr, 0, sizeof(sockAddr));

	int nSockAddrLen = sizeof(sockAddr);
	BOOL bResult = GetPeerName((SOCKADDR*)&sockAddr, &nSockAddrLen);
	if (bResult)
	{
		rPeerPort = ntohs(sockAddr.sin_port);
		rPeerAddress = inet_ntoa(sockAddr.sin_addr);
	}
	return bResult;
}
#endif //AFX

BOOL CAsyncSocketEx::GetPeerName( SOCKADDR* lpSockAddr, int* lpSockAddrLen )
{
#if defined(WIN32)
	if (m_pFirstLayer)
		return m_pFirstLayer->GetPeerName(lpSockAddr, lpSockAddrLen);
#endif
	if ( !getpeername(m_hSocket, lpSockAddr, lpSockAddrLen) )
		return TRUE;
	else
		return FALSE;
}

#ifdef _AFX
BOOL CAsyncSocketEx::GetSockName(CString& rSocketAddress, UINT& rSocketPort)
{
	SOCKADDR_IN sockAddr;
	memset(&sockAddr, 0, sizeof(sockAddr));

	int nSockAddrLen = sizeof(sockAddr);
	BOOL bResult = GetSockName((SOCKADDR*)&sockAddr, &nSockAddrLen);
	if (bResult)
	{
		rSocketPort = ntohs(sockAddr.sin_port);
		rSocketAddress = inet_ntoa(sockAddr.sin_addr);
	}
	return bResult;
}
#endif

BOOL CAsyncSocketEx::GetSockName( SOCKADDR* lpSockAddr, int* lpSockAddrLen )
{
	if ( !getsockname(m_hSocket, lpSockAddr, lpSockAddrLen) )
		return TRUE;
	else
		return FALSE;
}

BOOL CAsyncSocketEx::ShutDown( int nHow /*=sends*/ )
{
    int rc = SOCKET_ERROR;
#if defined(WIN32)
	if (m_pFirstLayer)
	{
		return m_pFirstLayer->ShutDown();
	}
	else
	{
#endif
		NET_VERIFY(shutdown(m_hSocket, nHow));
        return (SOCKET_ERROR != rc);
#if defined(WIN32)
	}
#endif
}

SOCKET CAsyncSocketEx::Detach()
{
	SOCKET socket=m_hSocket;

	TerminateAsyncSelect();
	sKillSocket( m_nSMT_Index, this );	

	CSocketHelperWnd::sRemoveConnectionSocketFromMap(this);

	m_nSMT_Index = -1;
	m_hSocket = INVALID_SOCKET;
	m_bNotifyClosed = false;
	return socket;
}

BOOL CAsyncSocketEx::Attach( SOCKET hSocket, long lEvent /*= FD_READ | FD_WRITE | FD_OOB | FD_ACCEPT | FD_CONNECT | FD_CLOSE*/ )
{
	if (hSocket==INVALID_SOCKET || !hSocket)
		return FALSE;

    int rc = SOCKET_ERROR;
	m_hSocket = hSocket;	

	// Attach ̹ Ŀؼ(Accept)  SOCKET ̰ڴٴ ǹ̹Ƿ
	// Ŀؼ Ʈ ̹ ִٸ !
	CSocketHelperWnd::sRemoveConnectionSocketFromMap(this);

	if( !sAttachHandle( this ) )
	{
		NET_VERIFY(closesocket( hSocket ));

		m_hSocket = INVALID_SOCKET;
		return FALSE;
	}

	m_bNotifyClosed = false;

	return AsyncSelect(lEvent);
}

BOOL CAsyncSocketEx::AsyncSelect( long lEvent /*= FD_READ | FD_WRITE | FD_OOB | FD_ACCEPT | FD_CONNECT | FD_CLOSE*/ )
{
//	ASSERT(m_pLocalAsyncSocketExThreadData);
    int rc = SOCKET_ERROR;
	m_lEvent=lEvent;
#if defined(WIN32)
	if( m_pFirstLayer )
		lEvent = FD_READ | FD_WRITE | FD_OOB | FD_ACCEPT | FD_CONNECT | FD_CLOSE;
#endif
    
    NET_VERIFY(WSAAsyncSelect(m_hSocket, GetHelperWindowHandle(),  WM_SOCKETEX_NOTIFY + m_nSMT_Index, lEvent));
	return (SOCKET_ERROR  !=  rc);
}

BOOL CAsyncSocketEx::TerminateAsyncSelect()
{
    int rc = 0;
    NET_VERIFY(WSAAsyncSelect(m_hSocket, GetHelperWindowHandle(), 0, 0));
	return (SOCKET_ERROR != rc);
}

int CAsyncSocketEx::SendWSA(LPWSABUF lpBuffers,
							DWORD dwBufferCount,
							LPDWORD lpNumberOfBytesSent,
							DWORD dwFlags,
							LPWSAOVERLAPPED lpOverlapped,
							LPWSAOVERLAPPED_COMPLETION_ROUTINE lpCompletionRoutine)
{
    int rc = SOCKET_ERROR;
	NET_VERIFY(::WSASend( m_hSocket,
		lpBuffers,
		dwBufferCount,
		lpNumberOfBytesSent,
		dwFlags,
		lpOverlapped,
		lpCompletionRoutine));
    return rc;
}
	
BOOL CAsyncSocketEx::Listen( int nConnectionBacklog /*=5*/ )
{
    int rc = SOCKET_ERROR;
	BOOL bRes = FALSE;
#if defined(WIN32)
	if (m_pFirstLayer)
	{
		bRes = m_pFirstLayer->Listen(nConnectionBacklog);
	}
	else
#endif
    {
        NET_VERIFY(listen(m_hSocket, nConnectionBacklog));
        bRes = (SOCKET_ERROR != rc);
    }

	if( bRes )
		m_bIsTcpListenSocket = true;

	return bRes;
}

BOOL CAsyncSocketEx::Accept( CAsyncSocketEx& rConnectedSocket, SOCKADDR* lpSockAddr /*=NULL*/, int* lpSockAddrLen /*=NULL*/ )
{
    int rc = SOCKET_ERROR;
	rConnectedSocket.m_bRequestConnect = true;

	ASSERT(rConnectedSocket.m_hSocket == INVALID_SOCKET);
#if defined(WIN32)
	if (m_pFirstLayer)
	{
		return m_pFirstLayer->Accept(rConnectedSocket, lpSockAddr, lpSockAddrLen);
	}
	else
	{
#endif
		SOCKET hTemp = accept(m_hSocket, lpSockAddr, lpSockAddrLen);
		if (hTemp == INVALID_SOCKET)
		{
            NET_VERIFY(SOCKET_ERROR); //make sure ASSERT here
			return FALSE;
		}
		else
		{
			rConnectedSocket.m_hSocket = hTemp;
			if( !sAttachHandle( &rConnectedSocket ) )
			{
				rConnectedSocket.m_hSocket = INVALID_SOCKET;
				NET_VERIFY(closesocket( hTemp ));
				return FALSE;
			}
			
			rConnectedSocket.m_bNotifyClosed = false;
			rConnectedSocket.AsyncSelect();
		}
#if defined(WIN32)
	}
#endif
	return TRUE;
}

BOOL CAsyncSocketEx::TriggerEvent(short lEvent, short error)
{
	if (m_hSocket==INVALID_SOCKET)
		return FALSE;

	ASSERT(ms_pHelperWindow);
#if defined(WIN32)
	if (m_pFirstLayer)
	{
		CAsyncSocketExLayer::t_LayerNotifyMsg *pMsg=new CAsyncSocketExLayer::t_LayerNotifyMsg;
		pMsg->lEvent = MAKELPARAM( lEvent, error );   // loword - event, hiword - error
		pMsg->pLayer=0;
		BOOL res=PostMessage(GetHelperWindowHandle(), WM_USER, (WPARAM)this, (LPARAM)pMsg);
		if (!res)
			delete pMsg;
		return res;
	}
	else
#endif
		return PostMessage(GetHelperWindowHandle(), 
						   WM_SOCKETEX_NOTIFY + m_nSMT_Index, m_hSocket, MAKELPARAM( lEvent, error ) );		// loword - event, hiword - error

}

HWND CAsyncSocketEx::GetHelperWindowHandle()
{
//	if (!m_pLocalAsyncSocketExThreadData)
//		return 0;
	if (!ms_pHelperWindow)
		return 0;
	return ms_pHelperWindow->GetHwnd();
}
#if defined(WIN32)
BOOL CAsyncSocketEx::AddLayer(CAsyncSocketExLayer *pLayer)
{
	ASSERT(pLayer);
	if (m_hSocket!=INVALID_SOCKET)
		return FALSE;

	if (m_pFirstLayer)
	{
		ASSERT(m_pLastLayer);
		m_pLastLayer=m_pLastLayer->AddLayer(pLayer, this);
		return m_pLastLayer?TRUE:FALSE;
	}
	else
	{
		ASSERT(!m_pLastLayer);
		pLayer->Init(0, this);
		m_pFirstLayer=pLayer;
		m_pLastLayer=m_pFirstLayer;
	}

	return TRUE;
}

void CAsyncSocketEx::RemoveAllLayers()
{
	m_pFirstLayer=0;
	m_pLastLayer=0;
}

int CAsyncSocketEx::OnLayerCallback(const CAsyncSocketExLayer *pLayer, int nType, int nParam1, int nParam2)
{
	ASSERT(pLayer);
	return 1;
}
#endif
BOOL CAsyncSocketEx::GetSockOpt(int nOptionName, void* lpOptionValue, int* lpOptionLen, int nLevel /*=SOL_SOCKET*/)
{
    int rc = SOCKET_ERROR;
	NET_VERIFY(getsockopt(m_hSocket, nLevel, nOptionName, (LPSTR)lpOptionValue, lpOptionLen));
    return (SOCKET_ERROR != rc);
}

BOOL CAsyncSocketEx::SetSockOpt(int nOptionName, const void* lpOptionValue, int nOptionLen, int nLevel /*=SOL_SOCKET*/)
{
    int rc = SOCKET_ERROR;
	NET_VERIFY(setsockopt(m_hSocket, nLevel, nOptionName, (LPSTR)lpOptionValue, nOptionLen));
    return (SOCKET_ERROR != rc);
}

/////////////////////////////////////////////////////////////////////////
// 2003-07-26 翵 ߰
// FD_WRITE ̺Ʈ  ޱ ؼ.. .. WSAAsyncSelect()  θ̴.

void  CAsyncSocketEx::CheckOnSendEvent()
{
    int rc = SOCKET_ERROR;
	NET_VERIFY(WSAAsyncSelect(m_hSocket, GetHelperWindowHandle(),
		WM_SOCKETEX_NOTIFY + m_nSMT_Index, FD_READ | FD_WRITE | FD_CLOSE));
}

